from typing import Type

from .base import AbstractKey, KeyAlgorithm
from .ecdsa import ECDSAAlgorithm, ECDSAKey
from .eddsa import EdDSAAlgorithm, EdDSAKey
from .hmac import HMACAlgorithm, HMACKey
from .rsa import RSAAlgorithm, RSAKey


def get_algorithm(alg: str) -> Type[KeyAlgorithm]:
    """
    Get the appropriate algorithm manager for the given algorithm.

    Args:
        alg: The algorithm name (e.g., "HS256", "RS256", "ES256", "EdDSA")

    Returns:
        The appropriate algorithm manager class

    Raises:
        ValueError: If the algorithm is not supported
    """
    for algo in KeyAlgorithm.__subclasses__():
        if alg.upper() in algo.SUPPORTED_ALGORITHMS:
            return algo

    raise ValueError(f"Unsupported algorithm: {alg}")


__all__ = [
    "KeyAlgorithm",
    "AbstractKey",
    "HMACKey",
    "RSAKey",
    "ECDSAKey",
    "EdDSAKey",
    "HMACAlgorithm",
    "RSAAlgorithm",
    "ECDSAAlgorithm",
    "EdDSAAlgorithm",
    "get_algorithm",
]



import base64
from enum import Enum


class Algorithm(str, Enum):
    # HMAC with SHA-2
    HS256 = "HS256"  # HMAC with SHA-256
    HS384 = "HS384"  # HMAC with SHA-384
    HS512 = "HS512"  # HMAC with SHA-512

    # RSA with SHA-2
    RS256 = "RS256"  # RSA with SHA-256
    RS384 = "RS384"  # RSA with SHA-384
    RS512 = "RS512"  # RSA with SHA-512

    # ECDSA with SHA-2
    ES256 = "ES256"  # ECDSA with SHA-256
    ES384 = "ES384"  # ECDSA with SHA-384
    ES512 = "ES512"  # ECDSA with SHA-512

    # RSASSA-PSS with SHA-2
    PS256 = "PS256"  # RSASSA-PSS with SHA-256
    PS384 = "PS384"  # RSASSA-PSS with SHA-384
    PS512 = "PS512"  # RSASSA-PSS with SHA-512

    # EdDSA with SHA-2
    EdDSA = "EDDSA"  # EdDSA with SHA-512


def b64url_decode(data: str) -> bytes:
    padding = "=" * (-len(data) % 4)
    return base64.urlsafe_b64decode(data + padding)


def b64url_encode(data: bytes) -> str:
    """Base64url encode bytes."""
    return base64.urlsafe_b64encode(data).rstrip(b"=").decode()



import httpx
import pytest

from src.usso_jwt import exceptions, sign, verify
from src.usso_jwt.algorithms import AbstractKey


def test_fetch_jwk():
    jwks_url = "https://www.googleapis.com/oauth2/v3/certs"
    jwks = httpx.get(jwks_url).json()
    jwk = jwks["keys"][0]
    fetched_jwk = verify.fetch_jwk(jwks_url=jwks_url, kid=jwk["kid"])
    assert fetched_jwk is not None
    assert fetched_jwk["kid"] == jwk["kid"]
    assert fetched_jwk["alg"] == jwk["alg"]
    assert fetched_jwk["kty"] == jwk["kty"]
    assert fetched_jwk["n"] is not None
    assert fetched_jwk["e"] is not None


def test_invalid_signature(
    test_valid_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_valid_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    jwt = jwt[:-1]  # + chr(ord(jwt[-1]) + 1)
    with pytest.raises(exceptions.JWTInvalidFormatError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())


def test_expired_payload(
    test_expired_payload: dict, test_header: dict, test_key: AbstractKey
):
    jwt = sign.generate_jwt(
        header=test_header,
        payload=test_expired_payload,
        key=test_key.private_der(),
        alg=test_key.algorithm,
    )
    with pytest.raises(exceptions.JWTExpiredError):
        verify.verify_jwt(token=jwt, jwk=test_key.jwk())



"""JWT-related exceptions."""


class JWTError(Exception):
    """Base exception for JWT-related errors."""

    def __init__(self, message: str = None, *args):
        self.message = message
        super().__init__(message, *args)


class JWTExpiredError(JWTError):
    """Raised when a JWT has expired."""

    def __init__(self):
        self.message = "JWT has expired"
        super().__init__(self.message)


class JWTNotValidYetError(JWTError):
    """Raised when a JWT's 'nbf' claim indicates it's not valid yet."""

    def __init__(self):
        self.message = "JWT is not valid yet"
        super().__init__(self.message)


class JWTIssuedInFutureError(JWTError):
    """Raised when a JWT's 'iat' claim is in the future."""

    def __init__(self):
        self.message = "JWT is issued in the future"
        super().__init__(self.message)


class JWTInvalidSignatureError(JWTError):
    """Raised when a JWT's signature is invalid."""

    def __init__(self):
        self.message = "JWT signature is invalid"
        super().__init__(self.message)


class JWTInvalidFormatError(JWTError):
    """Raised when a JWT has an invalid format."""

    def __init__(self):
        self.message = "JWT has an invalid format"
        super().__init__(self.message)


class JWKNotFoundError(JWTError):
    """Raised when a JWK with the specified kid is not found."""

    def __init__(self):
        self.message = "JWK with the specified kid is not found"
        super().__init__(self.message)
